@@ -11,6 +11,7 @@
 """Tests concerning model groups validation"""
 import unittest
 import os.path
+import warnings
 from textwrap import dedent
 from typing import Any, Union, List, Optional
 
@@ -18,7 +19,8 @@ from xmlschema import XMLSchema10, XMLSchema11
 from xmlschema.exceptions import XMLSchemaValueError
 from xmlschema.validators.exceptions import XMLSchemaValidationError
 from xmlschema.validators.particles import ParticleMixin
-from xmlschema.validators.models import distinguishable_paths, ModelVisitor
+from xmlschema.validators.models import distinguishable_paths, ModelVisitor, \
+    sort_content, iter_collapsed_content
 from xmlschema.validators.groups import XsdGroup
 from xmlschema.validators.elements import XsdElement
 from xmlschema.testing import XsdValidatorTestCase
@@ -30,8 +32,9 @@ class ModelGroup(XsdGroup):
     def __init__(self, model: str, min_occurs: int = 1, max_occurs: Optional[int] = 1) -> None:
         ParticleMixin.__init__(self, min_occurs, max_occurs)
         if model not in {'sequence', 'choice', 'all'}:
-            raise XMLSchemaValueError("invalid model {!r} for a group".format(model))
+            raise XMLSchemaValueError(f"invalid model {model!r} for a group")
         self._group: List[Union[ParticleMixin, 'ModelGroup']] = []
+        self.content = self._group
         self.model: str = model
 
     def __repr__(self) -> str:
@@ -115,7 +118,7 @@ class TestModelValidation(XsdValidatorTestCase):
 
         model = ModelVisitor(group)
         model.occurs[group[1]] = 1
-        self.assertListEqual(list(model.items), group[1:])
+        self.assertEqual(list(model.items), group[1:])
 
         group = ModelGroup('all')
         group.append(ParticleMixin())
@@ -124,7 +127,7 @@ class TestModelValidation(XsdValidatorTestCase):
 
         model = ModelVisitor(group)
         model.occurs[group[1]] = 1
-        self.assertListEqual(list(model.items), group[2:])
+        self.assertEqual(list(model.items), group[2:])
 
     # --- Vehicles schema ---
 
@@ -173,7 +176,9 @@ class TestModelValidation(XsdValidatorTestCase):
         self.assertIsNone(model.element)
 
         model = ModelVisitor(group)
-        self.check_advance_false(model, [(group[0], 0, [group[0]])])  # <not-a-car>
+        self.check_advance_false(
+            model, [(group[0], 0, [group[0]]), (group, 0, [group[0]])]
+        )  # <not-a-car>
         self.assertIsNone(model.element)
 
     def test_person_type_model(self):
@@ -212,7 +217,7 @@ class TestModelValidation(XsdValidatorTestCase):
           </xs:choice>
         </xs:group>
         """
-        group = XMLSchema10.meta_schema.groups['simpleDerivation']
+        group = self.schema_class.meta_schema.groups['simpleDerivation']
 
         model = ModelVisitor(group)
         self.check_advance_true(model)     # <restriction> matches
@@ -542,7 +547,9 @@ class TestModelValidation(XsdValidatorTestCase):
 
         model = ModelVisitor(group)
         self.assertEqual(model.element, group[0][0])
-        self.check_stop(model, [(group[0][0], 0, [group[0][0]])])
+        self.check_stop(
+            model, [(group[0][0], 0, [group[0][0]]), (group, 0, [group[0][0]])]
+        )
 
         group = self.models_schema.types['complexType7_emptiable'].content
 
@@ -620,35 +627,35 @@ class TestModelValidation(XsdValidatorTestCase):
                     <xs:choice>
                         <xs:any maxOccurs="2" processContents="lax"/>
                     </xs:choice>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="a2">
                 <xs:complexType>
                     <xs:choice>
                         <xs:any maxOccurs="2" processContents="strict"/>
                     </xs:choice>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="a3">
                 <xs:complexType>
                     <xs:sequence>
                         <xs:any maxOccurs="2" processContents="lax"/>
                     </xs:sequence>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="a4">
                 <xs:complexType>
                     <xs:choice>
                         <xs:element name="b" maxOccurs="2"/>
                     </xs:choice>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="a5">
                 <xs:complexType>
                     <xs:sequence>
                         <xs:element name="b" maxOccurs="2"/>
                     </xs:sequence>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="b"/>
         </xs:schema>""")
@@ -985,33 +992,24 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
 
         self.assertListEqual(
-            model.sort_content([('B2', 10), ('B1', 'abc'), ('B3', True)], restart=False),
+            sort_content([('B2', 10), ('B1', 'abc'), ('B3', True)], group),
             [('B1', 'abc'), ('B2', 10), ('B3', True)]
         )
         self.assertListEqual(
-            model.sort_content([('B2', 10), ('B1', 'abc'), ('B3', True)]),
+            sort_content([('B3', True), ('B2', 10), ('B1', 'abc')], group),
             [('B1', 'abc'), ('B2', 10), ('B3', True)]
         )
         self.assertListEqual(
-            model.sort_content([('B2', 10), ('B1', 'abc'), ('B3', True)], restart=False),
-            [('B2', 10), ('B1', 'abc'), ('B3', True)]
-        )
-
-        self.assertListEqual(
-            model.sort_content([('B3', True), ('B2', 10), ('B1', 'abc')]),
-            [('B1', 'abc'), ('B2', 10), ('B3', True)]
-        )
-        self.assertListEqual(
-            model.sort_content([('B2', 10), ('B4', None), ('B1', 'abc'), ('B3', True)]),
+            sort_content([('B2', 10), ('B4', None), ('B1', 'abc'), ('B3', True)], group),
             [('B1', 'abc'), ('B2', 10), ('B3', True), ('B4', None)]
         )
 
         content = [('B2', 10), ('B4', None), ('B1', 'abc'), (1, 'hello'), ('B3', True)]
         self.assertListEqual(
-            model.sort_content(content),
+            sort_content(content, group),
             [(1, 'hello'), ('B1', 'abc'), ('B2', 10), ('B3', True), ('B4', None)]
         )
 
@@ -1019,7 +1017,7 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             (2, 'world!'), ('B2', 10), ('B4', None), ('B1', 'abc'), (1, 'hello'), ('B3', True)
         ]
         self.assertListEqual(
-            model.sort_content(content),
+            sort_content(content, group),
             [(1, 'hello'), ('B1', 'abc'), (2, 'world!'), ('B2', 10), ('B3', True), ('B4', None)]
         )
 
@@ -1028,7 +1026,7 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             (5, 'five'), (4, 'four'), (2, 'two'), (3, 'three'), (1, 'one')
         ]
         self.assertListEqual(
-            model.sort_content(content),
+            sort_content(content, group),
             [(1, 'one'), ('B1', 'abc'), (2, 'two'), ('B2', 10), (3, 'three'),
              ('B3', True), (4, 'four'), ('B4', None), (5, 'five'), (6, 'six')]
         )
@@ -1036,26 +1034,30 @@ class TestModelBasedSorting(XsdValidatorTestCase):
         # With a dict-type argument
         content = dict([('B2', [10]), ('B1', ['abc']), ('B3', [True])])
         self.assertListEqual(
-            model.sort_content(content), [('B1', 'abc'), ('B2', 10), ('B3', True)]
+            sort_content(content, group), [('B1', 'abc'), ('B2', 10), ('B3', True)]
         )
         content = dict([('B2', [10]), ('B1', ['abc']), ('B3', [True]), (1, 'hello')])
         self.assertListEqual(
-            model.sort_content(content), [(1, 'hello'), ('B1', 'abc'), ('B2', 10), ('B3', True)]
+            sort_content(content, group),
+            [(1, 'hello'), ('B1', 'abc'), ('B2', 10), ('B3', True)]
         )
 
         # With partial content
-        self.assertListEqual(model.sort_content([]), [])
-        self.assertListEqual(model.sort_content([('B1', 'abc')]), [('B1', 'abc')])
-        self.assertListEqual(model.sort_content([('B2', 10)]), [('B2', 10)])
-        self.assertListEqual(model.sort_content([('B3', True)]), [('B3', True)])
+        self.assertListEqual(sort_content([], group), [])
+        self.assertListEqual(sort_content([('B1', 'abc')], group), [('B1', 'abc')])
+        self.assertListEqual(sort_content([('B2', 10)], group), [('B2', 10)])
+        self.assertListEqual(sort_content([('B3', True)], group), [('B3', True)])
         self.assertListEqual(
-            model.sort_content([('B3', True), ('B1', 'abc')]), [('B1', 'abc'), ('B3', True)]
+            sort_content([('B3', True), ('B1', 'abc')], group),
+            [('B1', 'abc'), ('B3', True)]
         )
         self.assertListEqual(
-            model.sort_content([('B2', 10), ('B1', 'abc')]), [('B1', 'abc'), ('B2', 10)]
+            sort_content([('B2', 10), ('B1', 'abc')], group),
+            [('B1', 'abc'), ('B2', 10)]
         )
         self.assertListEqual(
-            model.sort_content([('B3', True), ('B2', 10)]), [('B2', 10), ('B3', True)]
+            sort_content([('B3', True), ('B2', 10)], group),
+            [('B2', 10), ('B3', True)]
         )
 
     def test_iter_collapsed_content_with_optional_elements(self):
@@ -1074,18 +1076,15 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
-
+        group = schema.types['A_type'].content
         content = [('B3', 10), ('B4', None), ('B5', True), ('B6', 'alpha'), ('B7', 20)]
-        model.restart()
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)), content
+            list(iter_collapsed_content(content, group)), content
         )
 
         content = [('B3', 10), ('B5', True), ('B6', 'alpha'), ('B7', 20)]  # Missing B4
-        model.restart()
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)), content
+            list(iter_collapsed_content(content, group)), content
         )
 
     def test_iter_collapsed_content_with_repeated_elements(self):
@@ -1104,27 +1103,19 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
 
         content = [
             ('B3', 10), ('B4', None), ('B5', True), ('B5', False), ('B6', 'alpha'), ('B7', 20)
         ]
-        self.assertListEqual(
-            list(model.iter_collapsed_content(content)), content
-        )
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B3', 10), ('B3', 11), ('B3', 12), ('B4', None), ('B5', True),
                    ('B5', False), ('B6', 'alpha'), ('B7', 20), ('B7', 30)]
-        model.restart()
-        self.assertListEqual(
-            list(model.iter_collapsed_content(content)), content
-        )
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B3', 10), ('B3', 11), ('B3', 12), ('B4', None), ('B5', True), ('B5', False)]
-        model.restart()
-        self.assertListEqual(
-            list(model.iter_collapsed_content(content)), content
-        )
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
     def test_iter_collapsed_content_with_repeated_groups(self):
         schema = self.get_schema("""
@@ -1137,38 +1128,33 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
 
         content = [('B1', 1), ('B1', 2), ('B2', 3), ('B2', 4)]
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)),
+            list(iter_collapsed_content(content, group)),
             [('B1', 1), ('B2', 3), ('B1', 2), ('B2', 4)]
         )
 
         # Model broken by unknown element at start
         content = [('X', None), ('B1', 1), ('B1', 2), ('B2', 3), ('B2', 4)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 1), ('X', None), ('B1', 2), ('B2', 3), ('B2', 4)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 1), ('B1', 2), ('X', None), ('B2', 3), ('B2', 4)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 1), ('B1', 2), ('B2', 3), ('X', None), ('B2', 4)]
-        model.restart()
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)),
+            list(iter_collapsed_content(content, group)),
             [('B1', 1), ('B2', 3), ('B1', 2), ('X', None), ('B2', 4)]
         )
 
         content = [('B1', 1), ('B1', 2), ('B2', 3), ('B2', 4), ('X', None)]
-        model.restart()
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)),
+            list(iter_collapsed_content(content, group)),
             [('B1', 1), ('B2', 3), ('B1', 2), ('B2', 4), ('X', None)]
         )
 
@@ -1184,34 +1170,86 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
 
         content = [('B1', 'abc'), ('B2', 10), ('B3', False)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B3', False), ('B1', 'abc'), ('B2', 10)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 'abc'), ('B3', False), ('B2', 10)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 'abc'), ('B1', 'def'), ('B2', 10), ('B3', False)]
-        model.restart()
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)),
+            list(iter_collapsed_content(content, group)),
             [('B1', 'abc'), ('B2', 10), ('B3', False), ('B1', 'def')]
         )
 
         content = [('B1', 'abc'), ('B2', 10), ('X', None)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('X', None), ('B1', 'abc'), ('B2', 10), ('B3', False)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
+
+    def test_deprecated_methods(self):
+        schema = self.get_schema("""
+            <xs:element name="A" type="A_type" />
+            <xs:complexType name="A_type">
+                <xs:sequence maxOccurs="10">
+                    <xs:element name="B1" minOccurs="0" />
+                    <xs:element name="B2" minOccurs="0" />
+                </xs:sequence>
+            </xs:complexType>
+            """)
+
+        group = schema.types['A_type'].content
+        model = ModelVisitor(group)
+        default_namespace = 'http://xmlschema.test/ns'
+        content = [('B1', 1), ('B1', 2), ('B2', 3)]
+
+        with warnings.catch_warnings(record=True) as ctx:
+            warnings.   simplefilter("always")
+            self.assertListEqual(
+                list(model.iter_collapsed_content(content)),
+                [('B1', 1), ('B2', 3), ('B1', 2)]
+            )
+            self.assertEqual(len(ctx), 1, "Expected one deprecation warning")
+            self.assertIn("use iter_collapsed_content()", str(ctx[0].message))
+            self.assertNotIn("Don't provide default_namespace", str(ctx[0].message))
+
+        with warnings.catch_warnings(record=True) as ctx:
+            warnings.simplefilter("always")
+            self.assertListEqual(
+                list(model.iter_collapsed_content(content, default_namespace)),
+                [('B1', 1), ('B2', 3), ('B1', 2)]
+            )
+            self.assertEqual(len(ctx), 1, "Expected one deprecation warning")
+            self.assertIn("use iter_collapsed_content()", str(ctx[0].message))
+            self.assertIn("Don't provide default_namespace", str(ctx[0].message))
+
+        content = [('B2', 1), ('B1', 2), ('B2', 3), ('B1', 4)]
+
+        with warnings.catch_warnings(record=True) as ctx:
+            warnings.   simplefilter("always")
+            self.assertListEqual(
+                list(model.iter_unordered_content(content)),
+                [('B1', 2), ('B2', 1), ('B1', 4), ('B2', 3)]
+            )
+            self.assertEqual(len(ctx), 1, "Expected one deprecation warning")
+            self.assertIn("use iter_unordered_content()", str(ctx[0].message))
+            self.assertNotIn("Don't provide default_namespace", str(ctx[0].message))
+
+        with warnings.catch_warnings(record=True) as ctx:
+            warnings.simplefilter("always")
+            self.assertListEqual(
+                list(model.iter_unordered_content(content, default_namespace)),
+                [('B1', 2), ('B2', 1), ('B1', 4), ('B2', 3)]
+            )
+            self.assertEqual(len(ctx), 1, "Expected one deprecation warning")
+            self.assertIn("use iter_unordered_content()", str(ctx[0].message))
+            self.assertIn("Don't provide default_namespace", str(ctx[0].message))
 
 
 class TestModelPaths(unittest.TestCase):
